# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence, Tuple

import flowtorch
import torch
import torch.distributions.constraints as constraints
from flowtorch.bijectors.fixed import Fixed
from flowtorch.bijectors.volume_preserving import VolumePreserving
from torch.distributions.utils import lazy_property


class Permute(Fixed, VolumePreserving):
    domain = constraints.real_vector
    codomain = constraints.real_vector

    # TODO: A new abstraction so can defer construction of permutation
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        permutation: Optional[torch.Tensor] = None
    ) -> None:
        super().__init__(params_fn, shape=shape, context_shape=context_shape)
        self.permutation = permutation

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        if self.permutation is None:
            self.permutation = torch.randperm(x.shape[-1])

        y = torch.index_select(x, -1, self.permutation)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        if self.permutation is None:
            self.permutation = torch.randperm(y.shape[-1])

        x = torch.index_select(y, -1, self.inv_permutation)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    @lazy_property
    def inv_permutation(self) -> Optional[torch.Tensor]:
        if self.permutation is None:
            return None

        result = torch.empty_like(self.permutation, dtype=torch.long)
        result[self.permutation] = torch.arange(
            self.permutation.size(0), dtype=torch.long, device=self.permutation.device
        )
        return result



# Copyright (c) Meta Platforms, Inc

"""
Generates MDX (Markdown + JSX, see https://mdxjs.com/) files and sidebar
information for the Docusaurus v2 website from the library components'
docstrings.

We have chosen to take this approach to integrate our API documentation
with Docusaurus because there is no pre-existing robust solution to use
Sphinx output with Docusaurus.

This script will be run by the "documentation" GitHub workflow on pushes
and pull requests to the main branch. It will function corrrectly from
any working directory.

"""

import errno
import importlib
import inspect
import os
import re
from inspect import isclass, isfunction, ismodule
from typing import Any

import toml
from flowtorch.docs import (
    generate_class_markdown,
    generate_function_markdown,
    generate_module_markdown,
    sparse_module_hierarchy,
    walk_packages,
)


def module_sidebar(mod_name, items):
    return f"{{\n  type: 'category',\n  label: '{mod_name}',\n  \
collapsed: {'true'},\
  items: [{', '.join(items)}],\n}}"


def fullname(key, item):
    return key + "." + item


def dfs(dict):
    sidebar_items = []
    for key, val in dict.items():
        if len(modules_and_symbols[key][1]) > 0:
            items = [f'"api/{symbol_to_article[key]}"'] + [
                f'"api/{symbol_to_article[fullname(key, item)]}"'
                for item, _ in modules_and_symbols[key][1]
            ]
        else:
            items = []

        if val != {}:
            items.extend(dfs(val))

        sidebar_items.append(module_sidebar(key, items))

    return sidebar_items


# Generate article markdown files
def generate_markdown(article_name: str, symbol_name: str, entity: Any) -> str:
    """
    TODO: Method that inputs an object, extracts signature/docstring,
    and formats as markdown
    TODO: Method that build index markdown for overview files
    The overview for the entire API is a special case
    """

    if symbol_name == "":
        header = """---
id: overview
sidebar_label: "Overview"
slug: "/api"
---

:::info

These API stubs are generated from Python via a custom script and will filled
out in the future.

:::

"""
        return header

    # Regular modules/functions
    item = {
        "id": article_name,
        "sidebar_label": "Overview" if ismodule(entity) else symbol_name.split(".")[-1],
        "slug": f"/api/{article_name}",
        "ref": entity,
    }

    header = f"""---
id: {item['id']}
sidebar_label: {item['sidebar_label']}
---"""

    # Convert symbol to MDX

    # Imports for custom styling components
    markdown = [
        """import { FontAwesomeIcon } from '@fortawesome/react-fontawesome'
import { faAngleDoubleRight } from '@fortawesome/free-solid-svg-icons'
import PythonClass from "@theme/PythonClass";
import PythonFunction from "@theme/PythonFunction";
import PythonMethod from "@theme/PythonMethod";
import PythonModule from "@theme/PythonModule";
import PythonNavbar from "@theme/PythonNavbar";
"""
    ]

    # Make URL
    entity_file = (
        entity.__file__ if ismodule(entity) else inspect.getmodule(entity).__file__
    )
    url = (
        config["settings"]["github"]
        + "flowtorch/"
        + entity_file[(len(main_path) + 1) :].replace("\\", "/")
    )

    # Make navigation bar
    markdown.append(f"<PythonNavbar url='{url}'>\n")
    navigation = []
    symbol_splits = symbol_name.split(".")
    for idx in range(len(symbol_splits)):
        partial_symbol_name = ".".join(symbol_splits[0 : (idx + 1)])
        if idx == len(symbol_splits) - 1:
            navigation.append(f"*{symbol_splits[idx]}*")
        elif partial_symbol_name in symbol_to_article:
            navigation.append(
                f"[{symbol_splits[idx]}](/api/{symbol_to_article[partial_symbol_name]})"
            )
        else:
            navigation.append(f"{symbol_splits[idx]}")

    markdown.append(
        ' <FontAwesomeIcon icon={faAngleDoubleRight} size="sm" /> '.join(navigation)
    )
    markdown.append("\n</PythonNavbar>\n")

    # Handle known symbol types
    if isclass(entity):
        markdown.append(generate_class_markdown(symbol_name, entity))
        return "\n".join([header] + markdown)

    elif ismodule(entity):
        markdown.append(generate_module_markdown(symbol_name, entity))
        return "\n".join([header] + markdown)

    # Signature for function
    elif isfunction(entity):
        markdown.append(generate_function_markdown(symbol_name, entity))
        return "\n".join([header] + markdown)

    # Unknown symbol type
    else:
        raise ValueError(f"Symbol {symbol_name} has unknown type {type(symbol_object)}")


def search_symbols(config):
    # Validate module name to document
    assert (
        "settings" in config
        and "search" in config["settings"]
        and (
            type(config["settings"]["search"]) is str
            or type(config["settings"]["search"]) is list
        )
    )

    # TODO: Try to import module, more validation, etc.

    # Construct regular expressions for includes and excludes
    # Default include/exclude rules
    patterns = {
        "include": {"modules": re.compile(r".+"), "symbols": re.compile(r".+")},
        "exclude": {"modules": re.compile(r""), "symbols": re.compile(r"")},
    }

    # Override rules based on configuration file
    if "filters" in config:
        filters = config["filters"]
        for clude, rules in filters.items():
            for rule, pattern in rules.items():
                if type(pattern) is list:
                    pattern = "|".join(pattern)
                patterns[clude][rule] = re.compile(pattern)

    # Read in all modules and symbols
    search = config["settings"]["search"]
    search = [search] if type(search) is str else search
    modules_and_symbols = {}
    for modname in set(search):
        modules_and_symbols = {**modules_and_symbols, **walk_packages(modname)}

    # Apply filtering
    # TODO: Would be slightly faster if we applied module filtering inside walk_packages
    tmp = {}
    for x, y in modules_and_symbols.items():
        if (
            patterns["include"]["modules"].fullmatch(x) is not None
            and patterns["exclude"]["modules"].fullmatch(x) is None
        ):

            new_y1 = [
                (a, b)
                for a, b in y[1]
                if patterns["include"]["symbols"].fullmatch(x + "." + a) is not None
                and patterns["exclude"]["symbols"].fullmatch(x + "." + a) is None
            ]

            tmp[x] = (y[0], new_y1)

    return tmp


def construct_article_list(modules_and_symbols):
    # Construct list of articles (converting symbols to lower-case and collating)
    # NOTE: Webservers and Windows machines can't seem to distinguish addresses by
    # case...
    articles = {}
    symbol_to_article = {}

    for mod_name, (module, symbols) in modules_and_symbols.items():
        if len(symbols):
            article_name = mod_name.lower()
            # Find a unique name
            if article_name in articles:
                suffix = 1
                while article_name + str(suffix) in articles:
                    suffix += 1
                article_name = article_name + str(suffix)

            articles[article_name] = (mod_name, module)
            symbol_to_article[mod_name] = article_name

            suffix = 0
            for symbol_name, symbol in symbols:
                full_name = mod_name + "." + symbol_name
                article_name = full_name.lower()

                # Find a unique name
                if article_name in articles:
                    suffix += 1
                    while article_name + str(suffix) in articles:
                        suffix += 1
                    article_name = article_name + str(suffix)

                articles[article_name] = (full_name, symbol)
                symbol_to_article[full_name] = article_name

    return articles, symbol_to_article


if __name__ == "__main__":
    # Load and validate configuration file
    import flowtorch

    config_path = os.path.join(flowtorch.__path__[0], "../website/documentation.toml")
    config = toml.load(config_path)

    modules_and_symbols = search_symbols(config)
    articles, symbol_to_article = construct_article_list(modules_and_symbols)

    # Generate sidebar
    # Build hierarchy of modules
    hierarchy = sparse_module_hierarchy(modules_and_symbols.keys())

    # Create directories if they don't exist
    search = config["settings"]["search"]
    search = [search] if type(search) is str else search
    main_module = importlib.import_module(search[0])
    main_path = main_module.__path__[0]
    sidebar_path = os.path.join(main_path, config["paths"]["sidebar"])
    markdown_path = os.path.join(main_path, config["paths"]["markdown"])

    def create_paths(path: str) -> None:
        try:
            os.makedirs(path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    create_paths(sidebar_path)
    create_paths(markdown_path)

    with open(
        os.path.join(
            os.path.join(
                main_path,
                config["paths"]["sidebar"],
                config["paths"]["sidebar_filename"],
            )
        ),
        "w",
    ) as file:
        print("module.exports = [\n'api/overview',", file=file)
        print(",".join(dfs(hierarchy)), file=file)
        print("];", file=file)

    # For each article, convert the symbols to markdown, etc.

    # TODO: How to handle when there is a symbol called overview?
    # Maybe add a key for None instead of "overview"?
    articles["overview"] = ("", None)

    for article_name, (symbol_name, symbol_object) in articles.items():
        with open(
            os.path.join(
                os.path.join(
                    main_path, config["paths"]["markdown"], article_name + ".mdx"
                )
            ),
            "w",
        ) as file:
            print(
                generate_markdown(article_name, symbol_name, symbol_object), file=file
            )



# Copyright (c) Meta Platforms, Inc
from typing import Any, Optional, Iterator, Type, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from flowtorch.bijectors.base import Bijector

from torch import Tensor


class BijectiveTensor(Tensor):
    def __repr__(self) -> str:
        r_str = (
            super(BijectiveTensor, self)
            .__repr__()
            .replace("tensor", "bijective_tensor")
        )
        return r_str

    def register(
        self,
        input: Tensor,
        output: Tensor,
        context: Optional[Tensor],
        bijector: "Bijector",
        log_detJ: Optional[Tensor],
        mode: str,
    ) -> "BijectiveTensor":
        self._input = input
        self._output = output
        self._context = context
        self._bijector = bijector
        self._log_detJ = log_detJ
        self._mode = mode

        if not (self.from_forward() or self.from_inverse()):
            raise RuntimeError(
                f"BijectiveTensor mode must be either `'forward'` \
or `'inverse'`. got {self._mode}"
            )

        return self

    @classmethod
    def __torch_function__(
        cls: Type["BijectiveTensor"],
        func: Any,
        types: Any,
        args: Any = (),
        kwargs: Any = None,
    ) -> Union[Any, Tensor]:
        if kwargs is None:
            kwargs = {}
        # we don't want to create a new BijectiveTensor when summing,
        # calling zeros_like etc.
        types = tuple(Tensor if _type is BijectiveTensor else _type for _type in types)
        return Tensor.__torch_function__(func, types, args, kwargs)

    def check_bijector(self, bijector: "Bijector") -> bool:
        is_bijector = bijector in tuple(self.bijectors())
        return is_bijector

    def bijectors(self) -> Iterator["Bijector"]:
        yield self._bijector
        for parent in self.parents():
            if isinstance(parent, BijectiveTensor):
                yield parent._bijector

    def get_parent_from_bijector(self, bijector: "Bijector") -> Tensor:
        if self._bijector is bijector:
            return self.parent
        for parent in self.parents():
            if not isinstance(parent, BijectiveTensor):
                break
            if parent._bijector is bijector:
                return parent.parent
        raise RuntimeError("bijector not found in flow")

    def check_context(self, context: Optional[Tensor]) -> bool:
        return self._context is context

    def from_forward(self) -> bool:
        return self._mode == "forward"

    def from_inverse(self) -> bool:
        return self._mode == "inverse"

    def detach_from_flow(self) -> Tensor:
        detached_tensor = self._output if self.from_forward() else self._input
        if isinstance(detached_tensor, BijectiveTensor):
            raise RuntimeError("the detached tensor is an instance of BijectiveTensor.")
        return detached_tensor

    def has_ancestor(self, tensor: Tensor) -> bool:
        if tensor is self:
            return False  # self is no parent of self
        elif self.from_forward() and self._input is tensor:
            return True
        elif self.from_inverse() and self._output is tensor:
            return True
        elif self.from_forward() and isinstance(self._input, BijectiveTensor):
            return self._input.has_ancestor(tensor)
        elif self.from_inverse() and isinstance(self._output, BijectiveTensor):
            return self._output.has_ancestor(tensor)
        else:
            return False

    @property
    def log_detJ(self) -> Optional[Tensor]:
        return self._log_detJ

    @property
    def parent(self) -> Tensor:
        if self.from_forward():
            return self._input
        else:
            return self._output

    def parents(self) -> Iterator[Tensor]:
        child: Union[Tensor, BijectiveTensor] = self
        while True:
            assert isinstance(child, BijectiveTensor)
            child = parent = child.parent
            yield parent
            if not isinstance(child, BijectiveTensor):
                break


def to_bijective_tensor(
    x: Tensor,
    y: Tensor,
    context: Optional[Tensor],
    bijector: "Bijector",
    log_detJ: Optional[Tensor],
    mode: str = "forward",
) -> BijectiveTensor:
    if mode == "inverse":
        x_bij = BijectiveTensor(x)
        x_bij.register(x, y, context, bijector, log_detJ, mode=mode)
        return x_bij
    elif mode == "forward":
        y_bij = BijectiveTensor(y)
        y_bij.register(x, y, context, bijector, log_detJ, mode=mode)
        return y_bij
    else:
        raise NotImplementedError(
            f"mode {mode} is not supported, must be one of 'forward' or 'inverse'."
        )



# Copyright (c) Meta Platforms, Inc

import importlib
import inspect
import pkgutil
from collections import OrderedDict
from inspect import isclass, isfunction, signature
from types import ModuleType
from typing import Any, Callable, Dict, Mapping, Optional, Sequence, Tuple


def get_decorators(function: Callable) -> Sequence[str]:
    """Returns list of decorators names

    Args:
        function (Callable): decorated method/function

    Return:
        List of decorators as strings

    Example:
        Given:

        @my_decorator
        @another_decorator
        def decorated_function():
            pass

        >>> get_decorators(decorated_function)
        ['@my_decorator', '@another_decorator']

    """
    source = inspect.getsource(function)
    index = source.find("def ")
    return [
        line.strip().split()[0]
        for line in source[:index].strip().splitlines()
        if line.strip()[0] == "@"
    ]


def generate_class_markdown(symbol_name: str, entity: Any) -> str:
    markdown = []

    # Parents (for classes, this is like signature)
    parents = []
    for b in entity.__bases__:
        parents.append(b.__module__ + "." + b.__name__)
    parents_str = ", ".join(parents)

    # Docstring
    # TODO: Parse docstring and extract short summary
    docstring = entity.__doc__ if entity.__doc__ is not None else "empty docstring"
    docstring = "\n".join(line.strip() for line in docstring.splitlines())

    # short_summary = "```short summary```\n"
    safe_name = symbol_name.replace("_", r"\_")

    # Create top section for class
    markdown.append("<PythonClass>\n")
    markdown.append(
        """<div className="doc-class-row">
<div className="doc-class-label"><span className="doc-symbol-label">class</span></div>
<div className="doc-class-signature">\n"""
    )
    markdown.append(
        f"""## <span className="doc-symbol-name">{safe_name}</span> {{#class}}"""
    )
    markdown.append(
        f"""<span className="doc-inherits-from">Inherits from: <span className=\
"doc-symbol-name">{parents_str}</span></span>\n"""
    )
    # markdown.append(short_summary)
    markdown.append("</div>\n</div>\n\n</PythonClass>\n")
    markdown.append(f"```\n{docstring}\n```\n")

    # Methods for class
    members = inspect.getmembers(entity, predicate=inspect.isroutine)
    members = [
        (n, obj) for n, obj in members if n == "__init__" or not n.startswith("_")
    ]
    members = [(n, obj) for n, obj in members if type(obj) not in ["method_descriptor"]]

    # member_strs = []
    for member_name, member_object in members:
        # Try to unwrap class method and fetch decorators
        # decorators = []
        try:
            if hasattr(member_object, "__wrapped__"):
                # decorators = get_decorators(member_object)
                member_object = member_object.__wrapped__
        except Exception:
            pass

        # TODO: Prepend decorators to method name
        # for d in decorators:
        #    member_strs.append(d)

        markdown.append("<PythonMethod>\n")
        markdown.append(
            """<div className="doc-method-row">
<div className="doc-method-label"><span className="doc-symbol-label">member</span></div>
<div className="doc-method-signature">\n"""
        )

        # Some built-ins don't have a signature and throw exception...
        try:
            member_signature = str(signature(member_object))
        except ValueError:
            member_signature = "()"

        safe_member_signature = member_signature.replace("<", "&#60;").replace(
            ">", "&#62;"
        )
        # safe_member_signature = safe_member_signature.replace("'", "\'")

        safe_member_name = member_name.replace("_", r"\_")
        safe_member_id = member_name.replace("_", "-")
        markdown.append(
            f"""###  <span className="doc-symbol-name">{safe_member_name}</span>\
 {{#{safe_member_id}}}\n"""
        )
        markdown.append(
            f"""<span className="doc-symbol-signature">{safe_member_signature}\
</span>\n"""
        )
        markdown.append("</div>\n</div>\n\n</PythonMethod>\n")

        member_docstring = (
            member_object.__doc__
            if member_object.__doc__ is not None
            else "<empty docstring>"
        )
        member_docstring = "\n".join(
            line.strip() for line in member_docstring.splitlines()
        )
        markdown.append(f"```\n{member_docstring}\n```\n")

    return "\n".join(markdown)


def generate_module_markdown(symbol_name: str, entity: Any) -> str:
    markdown = []

    # Docstring
    # TODO: Parse docstring and extract short summary
    docstring = entity.__doc__ if entity.__doc__ is not None else "empty docstring"
    docstring = "\n".join(line.strip() for line in docstring.splitlines())

    # short_summary = "```short summary```\n"
    safe_name = symbol_name.replace("_", r"\_")

    # Create top section for class
    markdown.append("<PythonModule>\n")
    markdown.append(
        """<div className="doc-module-row">
<div className="doc-module-label"><span className="doc-symbol-label">module</span></div>
<div className="doc-module-signature">\n"""
    )

    markdown.append(
        f"""## <span className="doc-symbol-name">{safe_name}</span> {{#module}}\n"""
    )
    # markdown.append(short_summary)
    markdown.append("</div>\n</div>\n\n</PythonModule>\n")
    markdown.append(f"```\n{docstring}\n```\n")

    return "\n".join(markdown)


def generate_function_markdown(symbol_name: str, entity: Any) -> str:
    markdown = []

    # Docstring
    # TODO: Parse docstring and extract short summary
    docstring = entity.__doc__ if entity.__doc__ is not None else "empty docstring"
    docstring = "\n".join(line.strip() for line in docstring.splitlines())
    # short_summary = "```short summary```\n"

    # DEBUG
    try:
        if hasattr(entity, "__wrapped__"):
            # decorators = get_decorators(entity)
            entity = entity.__wrapped__
    except Exception:
        pass

    # TODO: Prepend decorators to method name
    # for d in decorators:
    #    member_strs.append(d)

    # Some built-ins don't have a signature and throw exception...
    try:
        entity_signature = str(signature(entity))
    except ValueError:
        entity_signature = "()"

    safe_name = symbol_name.replace("_", r"\_")
    safe_signature = entity_signature.replace("<", "&lt;").replace(">", "&gt;")

    markdown.append("<PythonFunction>\n")
    markdown.append(
        """<div className="doc-function-row">
<div className="doc-function-label"><span className="doc-symbol-label">function\
</span></div>
<div className="doc-function-signature">\n"""
    )

    markdown.append(
        f"""## <span className="doc-symbol-name">{safe_name}</span> \
{{#function}}\n"""
    )
    markdown.append(
        f"""<span className="doc-symbol-signature">{safe_signature}</span>"""
    )
    # markdown.append(short_summary)
    markdown.append("</div>\n</div>\n\n</PythonFunction>\n")
    markdown.append(f"```\n{docstring}\n```\n")

    return "\n".join(markdown)


def documentable_symbols(module: ModuleType) -> Sequence[Tuple[str, Any]]:
    """
    Given a module object, returns a list of object name and values for documentable
    symbols (functions and classes defined in this module or a subclass)
    """
    return [
        (n, m)
        for n, m in inspect.getmembers(module, None)
        if isfunction(m) or (isclass(m) and m.__module__.startswith(module.__name__))
    ]


def walk_packages(
    modname: str, filter: Optional[Callable[[Any], bool]] = None
) -> Mapping[str, Tuple[ModuleType, Sequence[Tuple[str, Any]]]]:
    """
    Given a base module name, return a mapping from the name of all modules
    accessible under the base to a tuple of module and symbol objects.

    A symbol is represented by a tuple of the object name and value, and is
    either a function or a class accessible when the module is imported.

    """
    module = importlib.import_module(modname)
    modules = {modname: (module, documentable_symbols(module))}

    # NOTE: I use path of flowtorch rather than e.g. flowtorch.bijectors
    # to avoid circular imports
    path = module.__path__  # type: ignore

    # The followings line uncovered a bug that hasn't been fixed in mypy:
    # https://github.com/python/mypy/issues/1422
    for importer, this_modname, _ in pkgutil.walk_packages(
        path=path,  # type: ignore  # mypy issue #1422
        prefix=f"{module.__name__}.",
        onerror=lambda x: None,
    ):
        # Conditions required for mypy
        if importer is not None:
            if isinstance(importer, importlib.abc.MetaPathFinder):
                finder = importer.find_module(this_modname, None)
            elif isinstance(importer, importlib.abc.PathEntryFinder):
                finder = importer.find_module(this_modname)
        else:
            finder = None

        if finder is not None:
            module = finder.load_module(this_modname)

        else:
            raise Exception("Finder is none")

        if module is not None:
            # Get all classes and functions imported/defined in module
            modules[this_modname] = (module, documentable_symbols(module))

            del module
            del finder

        else:
            raise Exception("Module is none")

    return modules


def sparse_module_hierarchy(mod_names: Sequence[str]) -> Mapping[str, Any]:
    # Make list of modules to search and their hierarchy, pruning entries that
    # aren't in mod_names
    results: Dict[str, Any] = OrderedDict()
    this_dict = results

    for module in sorted(mod_names):
        submodules = module.split(".")

        # Navigate to the correct insertion place for this module
        for idx in range(0, len(submodules)):
            submodule = ".".join(submodules[0 : (idx + 1)])
            if submodule in this_dict:
                this_dict = this_dict[submodule]

        # Insert module if it doesn't exist already
        this_dict.setdefault(module, {})

    return results



# Copyright (c) Meta Platforms, Inc
from typing import Optional, Sequence, Iterator

import flowtorch.parameters
import torch
import torch.distributions
from flowtorch.bijectors.base import Bijector
from flowtorch.bijectors.bijective_tensor import to_bijective_tensor, BijectiveTensor
from flowtorch.bijectors.utils import is_record_flow_graph_enabled, requires_log_detJ
from torch.distributions.utils import _sum_rightmost


class Compose(Bijector):
    def __init__(
        self,
        bijectors: Sequence[flowtorch.Lazy],
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
    ):
        assert len(bijectors) > 0

        # Instantiate all bijectors, propagating shape information
        self.bijectors = []
        for bijector in bijectors:
            assert issubclass(bijector.cls, Bijector)

            self.bijectors.append(bijector(shape=shape))
            shape = self.bijectors[-1].forward_shape(shape)  # type: ignore

        self.domain = self.bijectors[0].domain  # type: ignore
        self.codomain = self.bijectors[-1].codomain  # type: ignore

        self._context_shape = context_shape

    def parameters(self) -> Iterator[torch.Tensor]:
        for b in self.bijectors:
            for param in b.parameters():  # type: ignore
                yield param

    def forward(
        self,
        x: torch.Tensor,
        context: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        log_detJ: Optional[torch.Tensor] = None
        x_temp = x
        for bijector in self.bijectors:
            y = bijector.forward(x_temp, context)  # type: ignore
            if is_record_flow_graph_enabled() and requires_log_detJ():
                if isinstance(y, BijectiveTensor) and y.from_forward():
                    _log_detJ = y._log_detJ
                elif isinstance(x_temp, BijectiveTensor) and x_temp.from_inverse():
                    _log_detJ = x_temp._log_detJ
                else:
                    raise RuntimeError(
                        "neither of x nor y contains the log-abs-det-jacobian"
                    )
                log_detJ = log_detJ + _log_detJ if log_detJ is not None else _log_detJ
            x_temp = y

        # TODO: Check that this doesn't contain bugs!
        if (
            is_record_flow_graph_enabled()
            and not isinstance(y, BijectiveTensor)
            and not (isinstance(x, BijectiveTensor) and y in set(x.parents()))
        ):
            # we exclude y that are bijective tensors for Compose
            y = to_bijective_tensor(x, x_temp, context, self, log_detJ, mode="forward")
        return y

    def inverse(
        self,
        y: torch.Tensor,
        x: Optional[torch.Tensor] = None,
        context: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        log_detJ: Optional[torch.Tensor] = None
        y_temp = y
        for bijector in reversed(self.bijectors):
            x = bijector.inverse(y_temp, context)  # type: ignore
            if is_record_flow_graph_enabled() and requires_log_detJ():
                if isinstance(y_temp, BijectiveTensor) and y_temp.from_forward():
                    _log_detJ = y_temp._log_detJ
                elif isinstance(x, BijectiveTensor) and x.from_inverse():
                    _log_detJ = x._log_detJ
                else:
                    raise RuntimeError(
                        "neither of x nor y contains the log-abs-det-jacobian"
                    )
                log_detJ = log_detJ + _log_detJ if log_detJ is not None else _log_detJ
            y_temp = x  # type: ignore

        # TODO: Check that this doesn't contain bugs!
        if (
            is_record_flow_graph_enabled()
            and not isinstance(x, BijectiveTensor)
            and not (isinstance(y, BijectiveTensor) and x in set(y.parents()))
        ):
            x = to_bijective_tensor(y_temp, y, context, self, log_detJ, mode="inverse")
        return x  # type: ignore

    def log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, context: torch.Tensor = None
    ) -> torch.Tensor:
        """
        Computes the log det jacobian `log |dy/dx|` given input and output.
        By default, assumes a volume preserving bijection.
        """
        ldj = _sum_rightmost(
            torch.zeros_like(y),
            self.domain.event_dim,
        )

        if isinstance(x, BijectiveTensor) and x.has_ancestor(y):
            # If x is a BijectiveTensor and has y as ancestor, then the
            # inversion flow.inverse(y) = x has already been computed and
            # we can recover the chain of parents instead of re-computing it.
            _use_cached_inverse = True
            parents = []
            while isinstance(x, BijectiveTensor) and x is not y:
                parents.append(x)
                x = x.parent
        else:
            _use_cached_inverse = False

        for bijector in reversed(self.bijectors):
            if not _use_cached_inverse:
                y_inv = bijector.inverse(y, context)  # type: ignore
            else:
                y_inv = parents.pop()
            ldj += bijector.log_abs_det_jacobian(y_inv, y, context)  # type: ignore
            y = y_inv
        return ldj

    def param_shapes(self, shape: torch.Size) -> Sequence[torch.Size]:
        """
        Given a base distribution, calculate the parameters for the transformation
        of that distribution under this bijector. By default, no parameters are
        set.
        """
        return []



# Copyright (c) Meta Platforms, Inc
from typing import Any, Optional

import flowtorch
import torch
import torch.distributions
from flowtorch.bijectors.base import Bijector
from flowtorch.parameters.tensor import Tensor


class Elementwise(Bijector):
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        **kwargs: Any
    ) -> None:
        if not params_fn:
            params_fn = Tensor()  # type: ignore

        assert params_fn is None or issubclass(params_fn.cls, Tensor)

        super().__init__(params_fn, shape=shape, context_shape=context_shape)



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence, Tuple

import torch
import torch.distributions.constraints as constraints
import torch.nn.functional as F
from flowtorch.bijectors.fixed import Fixed
from flowtorch.ops import clipped_sigmoid


class Sigmoid(Fixed):
    codomain = constraints.unit_interval

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        y = clipped_sigmoid(x)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        finfo = torch.finfo(y.dtype)
        y = y.clamp(min=finfo.tiny, max=1.0 - finfo.eps)
        x = y.log() - torch.log1p(-y)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        return -F.softplus(-x) - F.softplus(x)



# Copyright (c) Meta Platforms, Inc
import functools
from typing import Any, Callable, List, Sequence

_RECORD_FLOW = True


class _context_manager:
    def __init__(self, value: bool = True) -> None:
        self.value = value
        self.prev: List[bool] = []

    def __call__(self, func: Callable) -> Any:
        @functools.wraps(func)
        def decorate_context(*args: Any, **kwargs: Sequence[Any]) -> Any:
            with self:
                return func(*args, **kwargs)

        return decorate_context

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        pass


class set_record_flow_graph(_context_manager):
    def __enter__(self) -> None:
        global _RECORD_FLOW
        self.prev.append(_RECORD_FLOW)
        _RECORD_FLOW = self.value

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        global _RECORD_FLOW
        _RECORD_FLOW = self.prev.pop()


def is_record_flow_graph_enabled() -> bool:
    return _RECORD_FLOW


_REQUIRES_LOG_DETJ = True


class set_requires_log_detJ(_context_manager):
    def __enter__(self) -> None:
        global _REQUIRES_LOG_DETJ
        self.prev.append(_REQUIRES_LOG_DETJ)
        _REQUIRES_LOG_DETJ = self.value

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        global _REQUIRES_LOG_DETJ
        _REQUIRES_LOG_DETJ = self.prev.pop()


def requires_log_detJ() -> bool:
    return _REQUIRES_LOG_DETJ



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence, Tuple

import flowtorch
import torch
import torch.distributions.constraints as constraints
from flowtorch.bijectors.fixed import Fixed


class Power(Fixed):
    r"""
    Elementwise bijector via the mapping :math:`y = x^{\text{exponent}}`.
    """
    domain = constraints.positive
    codomain = constraints.positive

    # TODO: Tensor valued exponents and corresponding determination of event_dim
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        exponent: float = 2.0,
    ) -> None:
        super().__init__(params_fn, shape=shape, context_shape=context_shape)
        self.exponent = exponent

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        y = x.pow(self.exponent)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self,
        y: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]] = None,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        x = y.pow(1 / self.exponent)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    def _log_abs_det_jacobian(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
    ) -> torch.Tensor:
        return torch.abs(self.exponent * y / x).log()



# Copyright (c) Meta Platforms, Inc

# This implementation is adapted in part from:
# * https://github.com/tonyduan/normalizing-flows/blob/master/nf/flows.py;
# * https://github.com/hmdolatabadi/LRS_NF/blob/master/nde/transforms/
#       nonlinearities.py; and,
# * https://github.com/bayesiains/nsf/blob/master/nde/transforms/splines/
#       rational_quadratic.py
# under the MIT license.

from typing import Any, Optional, Sequence, Tuple

import flowtorch
import torch
import torch.nn.functional as F
from flowtorch.bijectors.base import Bijector
from flowtorch.ops import monotonic_rational_spline
from torch.distributions.utils import _sum_rightmost


class Spline(Bijector):
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        count_bins: int = 8,
        bound: float = 3.0,
        order: str = "linear"
    ) -> None:
        if order not in ["linear", "quadratic"]:
            raise ValueError(
                "Keyword argument 'order' must be one of ['linear', \
'quadratic'], but '{}' was found!".format(
                    order
                )
            )

        self.count_bins = count_bins
        self.bound = bound
        self.order = order

        super().__init__(params_fn, shape=shape, context_shape=context_shape)

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        y, log_detJ = self._op(x, params)
        return y, _sum_rightmost(log_detJ, self.domain.event_dim)

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        x_new, log_detJ = self._op(y, params, inverse=True)

        # TODO: Should I invert the sign of log_detJ?
        # TODO: A unit test that compares log_detJ from _forward and _inverse
        return x_new, _sum_rightmost(log_detJ, self.domain.event_dim)

    def _log_abs_det_jacobian(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
    ) -> torch.Tensor:
        _, log_detJ = self._op(x, params)
        return _sum_rightmost(log_detJ, self.domain.event_dim)

    def _op(
        self,
        input: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
        inverse: bool = False,
        **kwargs: Any
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        assert params is not None

        lambdas: Optional[torch.Tensor] = None
        if self.order == "linear":
            widths, heights, derivatives, lambdas = params
            lambdas = torch.sigmoid(lambdas)
        else:
            widths, heights, derivatives = params

        # Constrain parameters
        # TODO: Move to flowtorch.ops function?
        widths = F.softmax(widths, dim=-1)
        heights = F.softmax(heights, dim=-1)
        derivatives = F.softplus(derivatives)

        y, log_detJ = monotonic_rational_spline(
            input,
            widths,
            heights,
            derivatives,
            lambdas,
            bound=self.bound,
            inverse=inverse,
            **kwargs
        )
        return y, log_detJ

    def param_shapes(self, shape: torch.Size) -> Sequence[torch.Size]:
        s1 = torch.Size(shape + (self.count_bins,))
        s2 = torch.Size(shape + (self.count_bins - 1,))

        if self.order == "linear":
            return s1, s1, s2, s1
        else:
            return s1, s1, s2



# Copyright (c) Meta Platforms, Inc

from typing import Optional

import flowtorch
import flowtorch.parameters
import torch
from flowtorch.bijectors.autoregressive import Autoregressive
from flowtorch.bijectors.ops.affine import Affine as AffineOp


class AffineAutoregressive(AffineOp, Autoregressive):
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        log_scale_min_clip: float = -5.0,
        log_scale_max_clip: float = 3.0,
        sigmoid_bias: float = 2.0,
    ) -> None:
        super().__init__(
            params_fn,
            shape=shape,
            context_shape=context_shape,
        )
        self.log_scale_min_clip = log_scale_min_clip
        self.log_scale_max_clip = log_scale_max_clip
        self.sigmoid_bias = sigmoid_bias



# Copyright (c) Meta Platforms, Inc

import warnings
from typing import Callable, Optional, Sequence

import torch
import torch.nn as nn
from flowtorch.nn.made import create_mask, MaskedLinear
from flowtorch.parameters.base import Parameters


class DenseAutoregressive(Parameters):
    autoregressive = True

    def __init__(
        self,
        param_shapes: Sequence[torch.Size],
        input_shape: torch.Size,
        context_shape: Optional[torch.Size],
        *,
        hidden_dims: Sequence[int] = (256, 256),
        nonlinearity: Callable[[], nn.Module] = nn.ReLU,
        permutation: Optional[torch.LongTensor] = None,
        skip_connections: bool = False,
    ) -> None:
        super().__init__(param_shapes, input_shape, context_shape)

        # Check consistency of input_shape with param_shapes
        # We need each param_shapes to match input_shape in
        # its leftmost dimensions
        for s in param_shapes:
            assert len(s) >= len(input_shape) and s[: len(input_shape)] == input_shape

        self.hidden_dims = hidden_dims
        self.nonlinearity = nonlinearity
        self.skip_connections = skip_connections
        self._build(input_shape, param_shapes, context_shape, permutation)

    def _build(
        self,
        input_shape: torch.Size,
        param_shapes: Sequence[torch.Size],
        context_shape: Optional[torch.Size],
        permutation: Optional[torch.LongTensor],
    ) -> None:
        # Work out flattened input and output shapes
        param_shapes_ = list(param_shapes)
        input_dims = int(torch.sum(torch.tensor(input_shape)).int().item())
        if input_dims == 0:
            input_dims = 1  # scalars represented by torch.Size([])
        if permutation is None:
            # By default set a random permutation of variables, which is
            # important for performance with multiple steps
            permutation = torch.LongTensor(
                torch.randperm(input_dims, device="cpu").to(
                    torch.LongTensor((1,)).device
                )
            )
        else:
            # The permutation is chosen by the user
            permutation = torch.LongTensor(permutation)

        self.param_dims = [
            int(max(torch.prod(torch.tensor(s[len(input_shape) :])).item(), 1))
            for s in param_shapes_
        ]

        self.output_multiplier = sum(self.param_dims)

        if input_dims == 1:
            warnings.warn(
                "DenseAutoregressive input_dim = 1. "
                "Consider using an affine transformation instead."
            )

        # Calculate the indices on the output corresponding to each parameter
        # TODO: Is this logic correct???
        # ends = torch.cumsum(
        #    torch.tensor(
        #        [max(torch.prod(torch.tensor(s)).item(), 1) for s in param_shapes_]
        #    ),
        #    dim=0,
        # )
        # starts = torch.cat((torch.zeros(1).type_as(ends), ends[:-1]))
        # self.param_slices = [slice(s.item(), e.item()) for s, e in zip(starts, ends)]

        # Hidden dimension must be not less than the input otherwise it isn't
        # possible to connect to the outputs correctly
        for h in self.hidden_dims:
            if h < input_dims:
                raise ValueError(
                    "Hidden dimension must not be less than input dimension."
                )

        # TODO: Check that the permutation is valid for the input dimension!
        # Implement ispermutation() that sorts permutation and checks whether it
        # has all integers from 0, 1, ..., self.input_dims - 1
        self.register_buffer("permutation", permutation)

        # Create masks
        hidden_dims = self.hidden_dims
        masks, mask_skip = create_mask(
            input_dim=input_dims,
            context_dim=0,  # context_dims,
            hidden_dims=hidden_dims,
            permutation=permutation,
            output_multiplier=self.output_multiplier,
        )

        # Create masked layers
        layers = [
            MaskedLinear(
                input_dims,  # + context_dims,
                hidden_dims[0],
                masks[0],
            ),
            self.nonlinearity(),
        ]
        for i in range(1, len(hidden_dims)):
            layers.extend(
                [
                    MaskedLinear(hidden_dims[i - 1], hidden_dims[i], masks[i]),
                    self.nonlinearity(),
                ]
            )
        layers.append(
            MaskedLinear(
                hidden_dims[-1],
                input_dims * self.output_multiplier,
                masks[-1],
            )
        )

        if self.skip_connections:
            layers.append(
                MaskedLinear(
                    input_dims,  # + context_dims,
                    input_dims * self.output_multiplier,
                    mask_skip,
                    bias=False,
                )
            )

        self.layers = nn.ModuleList(layers)

    def _forward(
        self,
        x: Optional[torch.Tensor] = None,
        context: Optional[torch.Tensor] = None,
    ) -> Optional[Sequence[torch.Tensor]]:
        assert x is not None

        # Flatten x
        batch_shape = x.shape[: len(x.shape) - len(self.input_shape)]
        if len(batch_shape) > 0:
            x = x.reshape(batch_shape + (-1,))

        if context is not None:
            # TODO: Fix the following!
            h = torch.cat([context.expand((x.shape[0], -1)), x], dim=-1)
        else:
            h = x

        for idx in range(len(self.layers) // 2):
            h = self.layers[2 * idx + 1](self.layers[2 * idx](h))
        h = self.layers[-1](h)

        # TODO: Get skip_layers working again!
        # if self.skip_layer is not None:
        #    h = h + self.skip_layer(x)

        # Shape the output
        # h ~ (batch_dims * input_dims, total_params_per_dim)
        h = h.reshape(-1, self.output_multiplier)

        # result ~ (batch_dims * input_dims, params_per_dim[0]), ...
        result = h.split(list(self.param_dims), dim=-1)

        # results ~ (batch_shape, param_shapes[0]), ...
        result = tuple(
            h_slice.view(batch_shape + p_shape)
            for h_slice, p_shape in zip(result, list(self.param_shapes))
        )
        return result



# Copyright (c) Meta Platforms, Inc

from typing import Any, cast, Optional, Sequence

import flowtorch
import flowtorch.parameters
import torch
import torch.distributions.constraints as constraints
from flowtorch.bijectors.base import Bijector
from flowtorch.bijectors.bijective_tensor import BijectiveTensor, to_bijective_tensor
from flowtorch.bijectors.utils import is_record_flow_graph_enabled
from flowtorch.parameters.dense_autoregressive import DenseAutoregressive


class Autoregressive(Bijector):
    # "Default" event shape is to operate on vectors
    domain = constraints.real_vector
    codomain = constraints.real_vector

    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        **kwargs: Any
    ) -> None:
        # Event shape is determined by `shape` argument
        self.domain = constraints.independent(constraints.real, len(shape))
        self.codomain = constraints.independent(constraints.real, len(shape))

        # currently only DenseAutoregressive has a `permutation` buffer
        if not params_fn:
            params_fn = DenseAutoregressive()  # type: ignore

        # TODO: Replace P.DenseAutoregressive with P.Autoregressive
        # In the future there will be other autoregressive parameter classes
        assert params_fn is not None and issubclass(params_fn.cls, DenseAutoregressive)

        super().__init__(params_fn, shape=shape, context_shape=context_shape)

    def inverse(
        self,
        y: torch.Tensor,
        x: Optional[torch.Tensor] = None,
        context: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # TODO: Allow that context can have a batch shape
        assert context is None  # or context.shape == (self._context_size,)
        assert self._params_fn is not None
        if self._check_bijective_y(y, context):
            assert isinstance(y, BijectiveTensor)
            return y.get_parent_from_bijector(self)
        x_new = torch.zeros_like(y)
        # NOTE: Inversion is an expensive operation that scales in the
        # dimension of the input
        permutation = (
            self._params_fn.permutation
        )  # TODO: type-safe named buffer (e.g. "permutation") access
        # TODO: Make permutation, inverse work for other event shapes
        log_detJ: Optional[torch.Tensor] = None
        for idx in cast(torch.LongTensor, permutation):
            _params = self._params_fn(x_new.clone(), context=context)
            x_temp, log_detJ = self._inverse(y, params=_params)
            x_new[..., idx] = x_temp[..., idx]
            # _log_detJ = out[1]
            # log_detJ = _log_detJ

        if is_record_flow_graph_enabled():
            x_new = to_bijective_tensor(
                x_new,
                y,
                context=context,
                bijector=self,
                mode="inverse",
                log_detJ=log_detJ,
            )
        return x_new

    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        raise NotImplementedError



# Copyright (c) Meta Platforms, Inc

"""
Warning: This file was generated by flowtorch/scripts/generate_imports.py
Do not modify or delete!

"""

import inspect
from typing import cast, List, Tuple

import torch
from flowtorch.bijectors.affine import Affine
from flowtorch.bijectors.affine_autoregressive import AffineAutoregressive
from flowtorch.bijectors.affine_fixed import AffineFixed
from flowtorch.bijectors.autoregressive import Autoregressive
from flowtorch.bijectors.base import Bijector
from flowtorch.bijectors.compose import Compose
from flowtorch.bijectors.elementwise import Elementwise
from flowtorch.bijectors.elu import ELU
from flowtorch.bijectors.exp import Exp
from flowtorch.bijectors.fixed import Fixed
from flowtorch.bijectors.leaky_relu import LeakyReLU
from flowtorch.bijectors.permute import Permute
from flowtorch.bijectors.power import Power
from flowtorch.bijectors.sigmoid import Sigmoid
from flowtorch.bijectors.softplus import Softplus
from flowtorch.bijectors.spline import Spline
from flowtorch.bijectors.spline_autoregressive import SplineAutoregressive
from flowtorch.bijectors.tanh import Tanh
from flowtorch.bijectors.volume_preserving import VolumePreserving

standard_bijectors = [
    ("Affine", Affine),
    ("AffineAutoregressive", AffineAutoregressive),
    ("AffineFixed", AffineFixed),
    ("ELU", ELU),
    ("Exp", Exp),
    ("LeakyReLU", LeakyReLU),
    ("Permute", Permute),
    ("Power", Power),
    ("Sigmoid", Sigmoid),
    ("Softplus", Softplus),
    ("Spline", Spline),
    ("SplineAutoregressive", SplineAutoregressive),
    ("Tanh", Tanh),
]

meta_bijectors = [
    ("Elementwise", Elementwise),
    ("Autoregressive", Autoregressive),
    ("Fixed", Fixed),
    ("Bijector", Bijector),
    ("Compose", Compose),
    ("VolumePreserving", VolumePreserving),
]


def isbijector(cls: type) -> bool:
    # A class must inherit from flowtorch.Bijector to be considered a valid bijector
    return issubclass(cls, Bijector)


def standard_bijector(cls: type) -> bool:
    # "Standard bijectors" are the ones we can perform standard automated tests upon
    return (
        inspect.isclass(cls)
        and isbijector(cls)
        and cls.__name__ not in [clx for clx, _ in meta_bijectors]
    )


# Determine invertible bijectors
invertible_bijectors = []
for bij_name, cls in standard_bijectors:
    # TODO: Use factored out version of the following
    # Define plan for flow
    event_dim = max(cls.domain.event_dim, 1)  # type: ignore
    event_shape = event_dim * [4]
    # base_dist = dist.Normal(torch.zeros(event_shape), torch.ones(event_shape))
    bij = cls(shape=torch.Size(event_shape))

    try:
        y = torch.randn(*bij.forward_shape(event_shape))
        bij.inverse(y)
    except NotImplementedError:
        pass
    else:
        invertible_bijectors.append((bij_name, cls))


__all__ = ["standard_bijectors", "meta_bijectors", "invertible_bijectors"] + [
    cls
    for cls, _ in cast(List[Tuple[str, Bijector]], meta_bijectors)
    + cast(List[Tuple[str, Bijector]], standard_bijectors)
]



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence, Tuple

import flowtorch
import torch
from flowtorch.bijectors.base import Bijector
from flowtorch.ops import clamp_preserve_gradients
from torch.distributions.utils import _sum_rightmost


class Affine(Bijector):
    r"""
    Affine mapping :math:`\mathbf{y} = \mu + \sigma \otimes \mathbf{x}` where
    $\mu$ and $\sigma$ are learnable parameters.

    """

    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        log_scale_min_clip: float = -5.0,
        log_scale_max_clip: float = 3.0,
        sigmoid_bias: float = 2.0,
    ) -> None:
        super().__init__(params_fn, shape=shape, context_shape=context_shape)
        self.log_scale_min_clip = log_scale_min_clip
        self.log_scale_max_clip = log_scale_max_clip
        self.sigmoid_bias = sigmoid_bias

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        assert params is not None

        mean, log_scale = params
        log_scale = clamp_preserve_gradients(
            log_scale, self.log_scale_min_clip, self.log_scale_max_clip
        )
        scale = torch.exp(log_scale)
        y = scale * x + mean
        return y, _sum_rightmost(log_scale, self.domain.event_dim)

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        assert params is not None

        mean, log_scale = params
        log_scale = clamp_preserve_gradients(
            log_scale, self.log_scale_min_clip, self.log_scale_max_clip
        )
        inverse_scale = torch.exp(-log_scale)
        x_new = (y - mean) * inverse_scale
        return x_new, _sum_rightmost(log_scale, self.domain.event_dim)

    def _log_abs_det_jacobian(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
    ) -> torch.Tensor:
        assert params is not None

        _, log_scale = params
        log_scale = clamp_preserve_gradients(
            log_scale, self.log_scale_min_clip, self.log_scale_max_clip
        )
        return _sum_rightmost(log_scale, self.domain.event_dim)

    def param_shapes(self, shape: torch.Size) -> Tuple[torch.Size, torch.Size]:
        # A mean and log variance for every dimension of the event shape
        return shape, shape



# Copyright (c) Meta Platforms, Inc

from typing import Optional

import flowtorch
import torch
from flowtorch.bijectors.elementwise import Elementwise
from flowtorch.bijectors.ops.spline import Spline as SplineOp


class Spline(SplineOp, Elementwise):
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        count_bins: int = 8,
        bound: float = 3.0,
        order: str = "linear"
    ) -> None:
        super().__init__(
            params_fn,
            shape=shape,
            context_shape=context_shape,
            count_bins=count_bins,
            bound=bound,
            order=order,
        )



# Copyright (c) Meta Platforms, Inc

import argparse
import os
import sys
from enum import Enum

from flowtorch.utils import copyright_header

lines_header = ["# " + ln + "\n" for ln in copyright_header.splitlines()]


class ReadState(Enum):
    EMPTY = 0
    COMMENT = 1
    TRIPLE_QUOTES = 2


def get_header(filename):
    state = ReadState.EMPTY
    header = []
    with open(filename, "r") as f:
        for line_idx, line in enumerate(f.readlines()):
            line = line.strip()
            # Finite state machine to read "header" of Python source
            # TODO: Can I write this much more compactly with regular expressions?
            if state is ReadState.EMPTY:
                if len(line) and line[0] == "#":
                    state = ReadState.COMMENT
                    header.append(line[1:].strip())
                    continue
                elif len(line) >= 3 and line[:3] == '"""':
                    state = ReadState.TRIPLE_QUOTES
                    header.append(line[3:].strip())
                    continue
                else:
                    # If the file doesn't begin with a comment we consider the
                    # header to be empty
                    return "\n".join(header).strip(), line_idx, state

            elif state is ReadState.COMMENT:
                if len(line) and line[0] == "#":
                    header.append(line[1:].strip())
                    continue
                else:
                    return "\n".join(header).strip(), line_idx, state

            elif state is ReadState.TRIPLE_QUOTES:
                if len(line) >= 3 and '"""' in line:
                    char_idx = line.find('"""')
                    header.append(line[:char_idx].strip())
                    return "\n".join(header).strip(), line_idx, state
                else:
                    header.append(line.strip())
                    continue

            else:
                raise RuntimeError("Invalid read state!")

    # Return error if triple quotes don't terminate
    if state is ReadState.TRIPLE_QUOTES:
        raise RuntimeError(f"Unterminated multi-line string in {f}")

    # If we get to here then the file is all header
    return "\n".join(header).strip(), line_idx + 1, state


def walk_source(paths):
    # Find all Python source files that are not Git ignored
    source_files = set()
    for path in paths:
        for root, _, files in os.walk(path):
            for name in files:
                full_name = os.path.join(root, name)
                if name.endswith(".py") and os.system(
                    f"git check-ignore -q {full_name}"
                ):
                    source_files.add(full_name)

    return sorted(source_files)


def print_results(count_changed, args):
    # Print results
    if count_changed == 0 and args.check:
        print(f"{count_changed} files would be left unchanged.")
    elif count_changed == len(source_files) and args.check:
        print(f"{count_changed} files would be changed.")
    elif args.check:
        print(
            f"""{count_changed} files would be changed and {len(source_files) /
                 - count_changed} files would be unchanged."""
        )
    elif count_changed:
        print(f"{count_changed} files fixed.")


if __name__ == "__main__":
    # Parse command line arguments
    # Example usage: python scripts/copyright_headers.py --check flowtorch tests scripts
    parser = argparse.ArgumentParser(
        description="Checks and adds the Meta Incubator copyright header"
    )
    parser.add_argument(
        "-c",
        "--check",
        action="store_true",
        help="just checks files and does not change any",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="prints extra information on files"
    )
    parser.add_argument(
        "paths", nargs="+", help="paths to search for Python source files"
    )
    args = parser.parse_args()

    source_files = walk_source(args.paths)

    # Loop over source files and get the "header"
    count_changed = 0
    for name in source_files:
        header, line_idx, state = get_header(name)

        # Replace if it's not equal, starts with empty space, or is not a comment
        if (
            header != copyright_header
            or line_idx != 1
            or not state == ReadState.COMMENT
        ):
            count_changed += 1
            if args.verbose:
                print(name)

            if not args.check:
                # Read the file
                with open(name, "r") as f:
                    lines = f.readlines()

                # Replace the header
                # TODO: Debug the following!
                if state == ReadState.TRIPLE_QUOTES:
                    after_quotes = lines[line_idx][
                        (lines[line_idx].find('"""') + 3) :
                    ].lstrip()
                    if after_quotes == "":
                        lines = lines[line_idx + 1 :]
                    elif after_quotes.startswith(";"):
                        lines = [after_quotes[1:].lstrip()] + lines[line_idx + 1 :]
                    else:
                        raise RuntimeError(
                            "Statements must be separated by newlines or semicolons"
                        )
                else:
                    lines = lines[line_idx:]

                lines = lines_header + lines
                filestring = "".join(lines)

                # Save back to disk
                with open(name, "w") as f:
                    f.write(filestring)

    print_results(count_changed, args)
    sys.exit(count_changed)



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence

import torch
import torch.nn as nn
from flowtorch.parameters.base import Parameters


class Tensor(Parameters):
    def __init__(
        self,
        param_shapes: Sequence[torch.Size],
        input_shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
    ) -> None:
        super().__init__(param_shapes, input_shape, context_shape)

        # TODO: Initialization strategies and constraints!
        self.params = nn.ParameterList(
            [nn.Parameter(torch.randn(shape) * 0.001) for shape in param_shapes]
        )

    def _forward(
        self, x: Optional[torch.Tensor] = None, context: Optional[torch.Tensor] = None
    ) -> Optional[Sequence[torch.Tensor]]:
        return list(self.params)



# Copyright (c) Meta Platforms, Inc

import inspect
from collections import OrderedDict
from typing import Any, Mapping, Tuple


# TODO: Move functions to flowtorch.utils?
def partial_signature(
    sig: inspect.Signature, *args: Any, **kwargs: Any
) -> Tuple[inspect.Signature, Mapping[str, Any]]:
    """
    Given an inspect.Signature object and a dictionary of (name, val) pairs,
    bind the names to the signature and return a new modified signature
    """
    bindings = dict(sig.bind_partial(*args, **kwargs).arguments)

    old_parameters = sig.parameters
    new_parameters = OrderedDict()

    for param_name in old_parameters:
        if param_name not in bindings:
            new_parameters[param_name] = old_parameters[param_name]

    bound_sig = sig.replace(parameters=list(new_parameters.values()))

    return bound_sig, bindings


def count_unbound(sig: inspect.Signature) -> int:
    return len(
        [p for p, v in sig.parameters.items() if v.default is inspect.Parameter.empty]
    )


class LazyMeta(type):
    def __call__(cls: Any, *args: Any, **kwargs: Any) -> Any:
        """
        Intercept instance creation
        """
        # Special behaviour for Lazy class
        if cls.__qualname__ == "Lazy":
            lazy_cls = args[0]
            args = args[1:]
        else:
            lazy_cls = cls

        # Remove first argument (i.e., self) from signature of class' initializer
        sig = inspect.signature(lazy_cls.__init__)
        new_parameters = OrderedDict(
            [(k, v) for idx, (k, v) in enumerate(sig.parameters.items()) if idx != 0]
        )
        sig = sig.replace(parameters=list(new_parameters.values()))

        # Attempt binding arguments to initializer
        bound_sig, bindings = partial_signature(sig, *args, **kwargs)

        # If there are no unbound arguments then instantiate class
        if not count_unbound(bound_sig):
            return type.__call__(lazy_cls, *args, **kwargs)

        # Otherwise, return Lazy instance
        else:
            return type.__call__(Lazy, lazy_cls, bindings, sig, bound_sig)


class Lazy(metaclass=LazyMeta):
    """
    Represents delayed instantiation of a class.
    """

    def __init__(
        self,
        cls: Any,
        bindings: Mapping[str, Any],
        sig: inspect.Signature,
        bound_sig: inspect.Signature,
    ):
        self.cls = cls
        self.bindings = bindings
        self.sig = sig
        self.bound_sig = bound_sig

    def __repr__(self) -> str:
        return f"Lazy(cls={self.cls.__name__}, bindings={self.bindings})"

    def __call__(self, *args: Any, **kwargs: Any) -> "Lazy":
        """
        Apply additional bindings
        """
        new_bindings = dict(self.bound_sig.bind_partial(*args, **kwargs).arguments)
        new_bindings.update(self.bindings)

        # Update args and kwargs
        new_args = []
        new_kwargs = {}
        for n, p in self.sig.parameters.items():
            if n in new_bindings:
                if p.kind == inspect.Parameter.POSITIONAL_ONLY:
                    new_args.append(new_bindings[n])

                else:
                    new_kwargs[n] = new_bindings[n]

        # Attempt object creation
        return Lazy(self.cls, *new_args, **new_kwargs)



# Copyright (c) Meta Platforms, Inc
from typing import Any, Dict, Optional, Union

import torch
import torch.distributions as dist
from torch.distributions import constraints
from torch.distributions.utils import _standard_normal


class NealsFunnel(dist.Distribution):
    """
    Neal's funnel.
    p(x,y) = N(y|0,3) N(x|0,exp(y/2))
    """

    support = constraints.real
    arg_constraints: Dict[str, dist.constraints.Constraint] = {}

    def __init__(self, validate_args: Any = None) -> None:
        d = 2
        batch_shape, event_shape = torch.Size([]), (d,)
        super(NealsFunnel, self).__init__(
            batch_shape, event_shape, validate_args=validate_args
        )

    def rsample(
        self,
        sample_shape: Union[torch.Tensor, torch.Size] = None,
        context: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        if not sample_shape:
            sample_shape = torch.Size()
        eps = _standard_normal(
            (sample_shape[0], 2), dtype=torch.float, device=torch.device("cpu")
        )
        z = torch.zeros(eps.shape)
        z[..., 1] = torch.tensor(3.0) * eps[..., 1]
        z[..., 0] = torch.exp(z[..., 1] / 2.0) * eps[..., 0]
        return z

    def log_prob(
        self, value: torch.Tensor, context: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        if self._validate_args:
            self._validate_sample(value)
        x = value[..., 0]
        y = value[..., 1]

        log_prob = dist.Normal(0, 3).log_prob(y)
        log_prob += dist.Normal(0, torch.exp(y / 2)).log_prob(x)

        return log_prob



# Copyright (c) Meta Platforms, Inc
from typing import Optional, Sequence

import flowtorch
import torch
import torch.distributions
from flowtorch.bijectors.base import Bijector


class Fixed(Bijector):
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
    ) -> None:
        # TODO: In the future, make Fixed actually mean that there is no autograd
        # through params
        super().__init__(params_fn, shape=shape, context_shape=context_shape)
        assert params_fn is None

    def param_shapes(self, shape: torch.Size) -> Sequence[torch.Size]:
        """
        Given a base distribution, calculate the parameters for the transformation
        of that distribution under this bijector. By default, no parameters are
        set.
        """
        # TODO: In the future, make Fixed actually mean that there is no autograd
        # through params
        return []



# Copyright (c) Meta Platforms, Inc

"""
Warning: This file was generated by flowtorch/scripts/generate_imports.py
Do not modify or delete!

"""

from flowtorch.parameters.base import Parameters
from flowtorch.parameters.dense_autoregressive import DenseAutoregressive
from flowtorch.parameters.tensor import Tensor

__all__ = ["Parameters", "DenseAutoregressive", "Tensor"]



# Copyright (c) Meta Platforms, Inc

"""
Generates imports files for bijectors, distributions, and parameters.
This script assumes that you have used `setup.py develop`.

"""

import errno
import io
import os

import black
import flowtorch
import torch
from flowtorch.utils import (
    classname,
    copyright_header,
    list_bijectors,
    list_distributions,
    list_parameters,
)

copyright_header = "".join(["# " + ln + "\n" for ln in copyright_header.splitlines()])

autogen_msg = """\"\"\"
Warning: This file was generated by flowtorch/scripts/generate_imports.py
Do not modify or delete!

\"\"\"
"""

bijectors_imports = """import inspect
from typing import cast, List, Tuple

import torch"""

bijectors_code = """def isbijector(cls: type) -> bool:
    # A class must inherit from flowtorch.Bijector to be considered a valid bijector
    return issubclass(cls, Bijector)


def standard_bijector(cls: type) -> bool:
    # "Standard bijectors" are the ones we can perform standard automated tests upon
    return (
        inspect.isclass(cls)
        and isbijector(cls)
        and cls.__name__ not in [clx for clx, _ in meta_bijectors]
    )

# Determine invertible bijectors
invertible_bijectors = []
for bij_name, cls in standard_bijectors:
    # TODO: Use factored out version of the following
    # Define plan for flow
    event_dim = max(cls.domain.event_dim, 1)  # type: ignore
    event_shape = event_dim * [4]
    # base_dist = dist.Normal(torch.zeros(event_shape), torch.ones(event_shape))
    bij = cls(shape=torch.Size(event_shape))

    try:
        y = torch.randn(*bij.forward_shape(event_shape))
        bij.inverse(y)
    except NotImplementedError:
        pass
    else:
        invertible_bijectors.append((bij_name, cls))


__all__ = ["standard_bijectors", "meta_bijectors", "invertible_bijectors"] + [
    cls
    for cls, _ in cast(List[Tuple[str, Bijector]], meta_bijectors)
    + cast(List[Tuple[str, Bijector]], standard_bijectors)
]"""

mode = black.FileMode()
fast = False


def generate_imports_plain(filename, classes):
    with io.StringIO() as file:
        # Sort classes by qualified name
        classes = sorted(classes, key=lambda tup: classname(tup[1]))

        print(copyright_header, file=file)
        print(autogen_msg, file=file)
        for s, cls in classes:
            print(f"from {cls.__module__} import {s}", file=file)

        print("", file=file)

        print("__all__ = [", file=file)
        all_list = ",\n\t".join([f'"{s}"' for s, _ in classes])
        print("\t", end="", file=file)
        print(all_list, file=file)
        print("]", end="", file=file)

        contents = file.getvalue()

    with open(filename, "w") as real_file:
        print(
            black.format_file_contents(contents, fast=fast, mode=mode),
            file=real_file,
            end="",
        )


def generate_imports_bijectors(filename):
    bij = list_bijectors()
    meta_bijectors = []
    standard_bijectors = []

    # Standard bijectors can be initialized with a shape, whereas
    # meta bijectors will throw a TypeError or require additional
    # keyword arguments (e.g., bij.Compose)
    # TODO: Refactor this into flowtorch.utils.ismetabijector
    for b in bij:
        try:
            cls = b[1]
            x = cls(shape=torch.Size([2] * cls.domain.event_dim))
        except TypeError:
            meta_bijectors.append(b)
        else:
            if isinstance(x, flowtorch.Lazy):
                meta_bijectors.append(b)
            else:
                standard_bijectors.append(b)

    with io.StringIO() as file:
        # Sort classes by qualified name
        classes = standard_bijectors + meta_bijectors
        classes = sorted(classes, key=lambda tup: classname(tup[1]))

        # Copyright header and warning message
        print(copyright_header, file=file)
        print(autogen_msg, file=file)

        # Non-FlowTorch imports
        print(bijectors_imports, file=file)

        # FlowTorch imports
        for s, cls in classes:
            print(f"from {cls.__module__} import {s}", file=file)
        print("", file=file)

        # Create lists of bijectors for each type
        meta_str = ",\n    ".join([f'("{b[0]}", {b[0]})' for b in meta_bijectors])
        standard_str = ",\n    ".join(
            [f'("{b[0]}", {b[0]})' for b in standard_bijectors]
        )

        print(
            f"""standard_bijectors = [
    {standard_str}
]
""",
            file=file,
        )

        print(
            f"""meta_bijectors = [
    {meta_str}
]
""",
            file=file,
        )

        # Rest of code
        print(bijectors_code, file=file)

        contents = file.getvalue()

    with open(filename, "w") as real_file:
        print(
            black.format_file_contents(contents, fast=fast, mode=mode),
            file=real_file,
            end="",
        )


if __name__ == "__main__":
    # Create module folders if they don't exist
    try:
        for m in ["distributions", "bijectors", "parameters"]:
            os.makedirs(os.path.join(flowtorch.__path__[0], m))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    bijectors_init = os.path.join(flowtorch.__path__[0], "bijectors/__init__.py")
    distributions_init = os.path.join(
        flowtorch.__path__[0], "distributions/__init__.py"
    )
    parameters_init = os.path.join(flowtorch.__path__[0], "parameters/__init__.py")

    generate_imports_bijectors(bijectors_init)
    generate_imports_plain(distributions_init, list_distributions())
    generate_imports_plain(parameters_init, list_parameters())



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Tuple

import torch
import torch.nn.functional as F

eps = 1e-8


def clamp_preserve_gradients(x: torch.Tensor, min: float, max: float) -> torch.Tensor:
    """
    This helper function clamps gradients but still passes through the
    gradient in clamped regions
    """
    return x + (x.clamp(min, max) - x).detach()


def clipped_sigmoid(x: torch.Tensor) -> torch.Tensor:
    finfo = torch.finfo(x.dtype)
    return torch.clamp(torch.sigmoid(x), min=finfo.tiny, max=1.0 - finfo.eps)


def softplus_inv(y: torch.Tensor) -> torch.Tensor:
    return y + y.neg().expm1().neg().log()


def _searchsorted(sorted_sequence: torch.Tensor, values: torch.Tensor) -> torch.Tensor:
    """
    Searches for which bin an input belongs to (in a way that is parallelizable and
    amenable to autodiff)
    TODO: Replace with torch.searchsorted once it is released
    """
    return torch.sum(values[..., None] >= sorted_sequence, dim=-1) - 1


def _select_bins(x: torch.Tensor, idx: torch.Tensor) -> torch.Tensor:
    """
    Performs gather to select the bin in the correct way on batched inputs
    """
    idx = idx.clamp(min=0, max=x.size(-1) - 1)

    # Broadcast dimensions of idx over x
    # idx ~ (batch_dims, input_dim, 1)
    # x ~ (context_batch_dims, input_dim, count_bins)
    # Note that by convention, the context variable batch dimensions must broadcast
    # over the input batch dimensions.
    if len(idx.shape) >= len(x.shape):
        x = x.reshape((1,) * (len(idx.shape) - len(x.shape)) + x.shape)
        x = x.expand(idx.shape[:-2] + (-1,) * 2)

    return x.gather(-1, idx).squeeze(-1)


def _calculate_knots(
    lengths: torch.Tensor, lower: float, upper: float
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Given a tensor of unscaled bin lengths that sum to 1, plus the lower and upper
    limits, returns the shifted and scaled lengths plus knot positions
    """

    # Cumulative widths gives x (y for inverse) position of knots
    knots = torch.cumsum(lengths, dim=-1)

    # Pad left of last dimension with 1 zero to compensate for dim lost to cumsum
    knots = F.pad(knots, pad=(1, 0), mode="constant", value=0.0)

    # Translate [0,1] knot points to [-B, B]
    knots = (upper - lower) * knots + lower

    # Convert the knot points back to lengths
    # NOTE: Are following two lines a necessary fix for accumulation (round-off) error?
    knots[..., 0] = lower
    knots[..., -1] = upper
    lengths = knots[..., 1:] - knots[..., :-1]

    return lengths, knots


def monotonic_rational_spline(
    inputs: torch.Tensor,
    widths: torch.Tensor,
    heights: torch.Tensor,
    derivatives: torch.Tensor,
    lambdas: Optional[torch.Tensor] = None,
    inverse: bool = False,
    bound: float = 3.0,
    min_bin_width: float = 1e-3,
    min_bin_height: float = 1e-3,
    min_derivative: float = 1e-3,
    min_lambda: float = 0.025,
    eps: float = 1e-6,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Calculating a monotonic rational spline (linear or quadratic) or its inverse,
    plus the log(abs(detJ)) required for normalizing flows.
    NOTE: I omit the docstring with parameter descriptions for this method since it
    is not considered "public" yet!
    """

    # Ensure bound is positive
    # NOTE: For simplicity, we apply the identity function outside [-B, B] X [-B, B]
    # rather than allowing arbitrary corners to the bounding box. If you want a
    # different bounding box you can apply an affine transform before and after the
    # input.
    assert bound > 0.0

    num_bins = widths.shape[-1]
    if min_bin_width * num_bins > 1.0:
        raise ValueError("Minimal bin width too large for the number of bins")
    if min_bin_height * num_bins > 1.0:
        raise ValueError("Minimal bin height too large for the number of bins")

    # inputs, inside_interval_mask, outside_interval_mask ~ (batch_dim, input_dim)
    left, right = -bound, bound
    bottom, top = -bound, bound
    inside_interval_mask = (inputs >= left) & (inputs <= right)
    outside_interval_mask = ~inside_interval_mask

    # outputs, logabsdet ~ (batch_dim, input_dim)
    outputs = torch.zeros_like(inputs)
    logabsdet = torch.zeros_like(inputs)

    # For numerical stability, put lower/upper limits on parameters. E.g. give
    # every bin min_bin_width, then add width fraction of remaining length.
    # NOTE: Do this here rather than higher up because we want everything to
    # ensure numerical stability within this function.
    widths = min_bin_width + (1.0 - min_bin_width * num_bins) * widths
    heights = min_bin_height + (1.0 - min_bin_height * num_bins) * heights
    derivatives = min_derivative + derivatives

    # Cumulative widths are x (y for inverse) position of knots
    # Similarly, cumulative heights are y (x for inverse) position of knots
    widths, cumwidths = _calculate_knots(widths, left, right)
    heights, cumheights = _calculate_knots(heights, bottom, top)

    # Pad left and right derivatives with fixed values at first and last knots
    # These are 1 since the function is the identity outside the bounding box
    # and the derivative is continuous.
    # NOTE: Not sure why this is 1.0 - min_derivative rather than 1.0. I've
    # copied this from original implementation
    derivatives = F.pad(
        derivatives, pad=(1, 1), mode="constant", value=1.0 - min_derivative
    )

    # Get the index of the bin that each input is in
    # bin_idx ~ (batch_dim, input_dim, 1)
    bin_idx = _searchsorted(
        cumheights + eps if inverse else cumwidths + eps, inputs
    ).unsqueeze(-1)

    # Select the value for the relevant bin for the variables
    # used in the main calculation
    input_widths = _select_bins(widths, bin_idx)
    input_cumwidths = _select_bins(cumwidths, bin_idx)
    input_cumheights = _select_bins(cumheights, bin_idx)
    input_delta = _select_bins(heights / widths, bin_idx)
    input_derivatives = _select_bins(derivatives, bin_idx)
    input_derivatives_plus_one = _select_bins(derivatives[..., 1:], bin_idx)
    input_heights = _select_bins(heights, bin_idx)

    # Calculate monotonic *linear* rational spline
    if lambdas is not None:
        lambdas = (1 - 2 * min_lambda) * lambdas + min_lambda
        input_lambdas = _select_bins(lambdas, bin_idx)

        # The weight, w_a, at the left-hand-side of each bin
        # We are free to choose w_a, so set it to 1
        wa = 1.0

        # The weight, w_b, at the right-hand-side of each bin
        # This turns out to be a multiple of the w_a
        # TODO: Should this be done in log space for numerical stability?
        wb = torch.sqrt(input_derivatives / input_derivatives_plus_one) * wa

        # The weight, w_c, at the division point of each bin
        # Recall that each bin is divided into two parts so we have enough
        # d.o.f. to fit spline
        wc = (
            input_lambdas * wa * input_derivatives
            + (1 - input_lambdas) * wb * input_derivatives_plus_one
        ) / input_delta

        # Calculate y coords of bins
        ya = input_cumheights
        yb = input_heights + input_cumheights
        yc = ((1.0 - input_lambdas) * wa * ya + input_lambdas * wb * yb) / (
            (1.0 - input_lambdas) * wa + input_lambdas * wb
        )

        if inverse:
            numerator = (input_lambdas * wa * (ya - inputs)) * (
                inputs <= yc
            ).float() + (
                (wc - input_lambdas * wb) * inputs + input_lambdas * wb * yb - wc * yc
            ) * (
                inputs > yc
            ).float()

            denominator = ((wc - wa) * inputs + wa * ya - wc * yc) * (
                inputs <= yc
            ).float() + ((wc - wb) * inputs + wb * yb - wc * yc) * (inputs > yc).float()

            theta = numerator / denominator

            outputs = theta * input_widths + input_cumwidths

            derivative_numerator = (
                wa * wc * input_lambdas * (yc - ya) * (inputs <= yc).float()
                + wb * wc * (1 - input_lambdas) * (yb - yc) * (inputs > yc).float()
            ) * input_widths

            logabsdet = torch.log(derivative_numerator) - 2 * torch.log(
                torch.abs(denominator)
            )

        else:
            theta = (inputs - input_cumwidths) / input_widths

            numerator = (wa * ya * (input_lambdas - theta) + wc * yc * theta) * (
                theta <= input_lambdas
            ).float() + (wc * yc * (1 - theta) + wb * yb * (theta - input_lambdas)) * (
                theta > input_lambdas
            ).float()

            denominator = (wa * (input_lambdas - theta) + wc * theta) * (
                theta <= input_lambdas
            ).float() + (wc * (1 - theta) + wb * (theta - input_lambdas)) * (
                theta > input_lambdas
            ).float()

            outputs = numerator / denominator

            derivative_numerator = (
                wa * wc * input_lambdas * (yc - ya) * (theta <= input_lambdas).float()
                + wb
                * wc
                * (1 - input_lambdas)
                * (yb - yc)
                * (theta > input_lambdas).float()
            ) / input_widths

            logabsdet = torch.log(derivative_numerator) - 2 * torch.log(
                torch.abs(denominator)
            )

    # Calculate monotonic *quadratic* rational spline
    else:
        if inverse:
            a = (inputs - input_cumheights) * (
                input_derivatives + input_derivatives_plus_one - 2 * input_delta
            ) + input_heights * (input_delta - input_derivatives)
            b = input_heights * input_derivatives - (inputs - input_cumheights) * (
                input_derivatives + input_derivatives_plus_one - 2 * input_delta
            )
            c = -input_delta * (inputs - input_cumheights)

            discriminant = b.pow(2) - 4 * a * c
            assert (discriminant >= 0).all()

            root = (2 * c) / (-b - torch.sqrt(discriminant))
            outputs = root * input_widths + input_cumwidths

            theta_one_minus_theta = root * (1 - root)
            denominator = input_delta + (
                (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
                * theta_one_minus_theta
            )
            derivative_numerator = input_delta.pow(2) * (
                input_derivatives_plus_one * root.pow(2)
                + 2 * input_delta * theta_one_minus_theta
                + input_derivatives * (1 - root).pow(2)
            )
            logabsdet = -(torch.log(derivative_numerator) - 2 * torch.log(denominator))

        else:
            theta = (inputs - input_cumwidths) / input_widths
            theta_one_minus_theta = theta * (1 - theta)

            numerator = input_heights * (
                input_delta * theta.pow(2) + input_derivatives * theta_one_minus_theta
            )
            denominator = input_delta + (
                (input_derivatives + input_derivatives_plus_one - 2 * input_delta)
                * theta_one_minus_theta
            )
            outputs = input_cumheights + numerator / denominator

            derivative_numerator = input_delta.pow(2) * (
                input_derivatives_plus_one * theta.pow(2)
                + 2 * input_delta * theta_one_minus_theta
                + input_derivatives * (1 - theta).pow(2)
            )
            logabsdet = torch.log(derivative_numerator) - 2 * torch.log(denominator)

    # Apply the identity function outside the bounding box
    outputs[outside_interval_mask] = inputs[outside_interval_mask]
    logabsdet[outside_interval_mask] = 0.0
    return outputs, logabsdet



# Copyright (c) Meta Platforms, Inc
import warnings
from typing import Optional, Sequence, Tuple, Union, Callable, Iterator

import flowtorch.parameters
import torch
import torch.distributions
from flowtorch.bijectors.bijective_tensor import to_bijective_tensor, BijectiveTensor
from flowtorch.bijectors.utils import is_record_flow_graph_enabled
from flowtorch.parameters import Parameters
from torch.distributions import constraints

ParamFnType = Callable[
    [Optional[torch.Tensor], Optional[torch.Tensor]], Optional[Sequence[torch.Tensor]]
]


class Bijector(metaclass=flowtorch.LazyMeta):
    codomain: constraints.Constraint = constraints.real
    domain: constraints.Constraint = constraints.real
    _shape: torch.Size
    _context_shape: Optional[torch.Size]
    _params_fn: Optional[Union[Parameters, torch.nn.ModuleList]] = None

    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
    ) -> None:
        # Prevent "meta bijectors" from being initialized
        # NOTE: We define a "standard bijector" as one that inherits from a
        # subclass of Bijector, hence why we need to test the length of the MRO
        if (
            self.__class__.__module__ == "flowtorch.bijectors.base"
            or len(self.__class__.__mro__) <= 3
        ):
            raise TypeError("Only standard bijectors can be initialized.")

        self._shape = shape
        self._context_shape = context_shape

        # Instantiate parameters (tensor, hypernets, etc.)
        if params_fn is not None:
            param_shapes = self.param_shapes(shape)
            self._params_fn = params_fn(  # type: ignore
                param_shapes, self._shape, self._context_shape
            )

    def parameters(self) -> Iterator[torch.Tensor]:
        assert self._params_fn is not None
        if hasattr(self._params_fn, "parameters"):
            for param in self._params_fn.parameters():
                yield param

    def _check_bijective_x(
        self, x: torch.Tensor, context: Optional[torch.Tensor]
    ) -> bool:
        return (
            isinstance(x, BijectiveTensor)
            and x.from_inverse()
            and x.check_bijector(self)
            and x.check_context(context)
        )

    def forward(
        self,
        x: torch.Tensor,
        context: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # TODO: Allow that context can have a batch shape
        assert context is None  # or context.shape == (self._context_size,)
        if self._check_bijective_x(x, context):
            assert isinstance(x, BijectiveTensor)
            return x.get_parent_from_bijector(self)

        params = self._params_fn(x, context) if self._params_fn is not None else None
        y, log_detJ = self._forward(x, params)
        if (
            is_record_flow_graph_enabled()
            and not isinstance(y, BijectiveTensor)
            and not (isinstance(x, BijectiveTensor) and y in set(x.parents()))
        ):
            # we exclude y that are bijective tensors for Compose
            y = to_bijective_tensor(x, y, context, self, log_detJ, mode="forward")
        return y

    def _forward(
        self,
        x: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Abstract method to compute forward transformation.
        """
        raise NotImplementedError

    def _check_bijective_y(
        self, y: torch.Tensor, context: Optional[torch.Tensor]
    ) -> bool:
        return (
            isinstance(y, BijectiveTensor)
            and y.from_forward()
            and y.check_bijector(self)
            and y.check_context(context)
        )

    def inverse(
        self,
        y: torch.Tensor,
        x: Optional[torch.Tensor] = None,
        context: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # TODO: Allow that context can have a batch shape
        assert context is None  # or context.shape == (self._context_size,)
        if self._check_bijective_y(y, context):
            assert isinstance(y, BijectiveTensor)
            return y.get_parent_from_bijector(self)

        # TODO: What to do in this line?
        params = self._params_fn(x, context) if self._params_fn is not None else None
        x, log_detJ = self._inverse(y, params)

        if (
            is_record_flow_graph_enabled()
            and not isinstance(x, BijectiveTensor)
            and not (isinstance(y, BijectiveTensor) and x in set(y.parents()))
        ):
            x = to_bijective_tensor(x, y, context, self, log_detJ, mode="inverse")
        return x

    def _inverse(
        self,
        y: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Abstract method to compute inverse transformation.
        """
        raise NotImplementedError

    def log_abs_det_jacobian(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        context: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Computes the log det jacobian `log |dy/dx|` given input and output.
        By default, assumes a volume preserving bijection.
        """
        # TODO: Allow that context can have a batch shape
        assert context is None  # or context.shape == (self._context_size,)
        ladj = None
        if (
            isinstance(y, BijectiveTensor)
            and y.from_forward()
            and y.check_bijector(self)
            and y.check_context(context)
        ):
            ladj = y.log_detJ
        elif (
            isinstance(x, BijectiveTensor)
            and x.from_inverse()
            and x.check_bijector(self)
            and x.check_context(context)
        ):
            ladj = x.log_detJ
        if ladj is None:
            if is_record_flow_graph_enabled():
                warnings.warn(
                    "Computing _log_abs_det_jacobian from values and not from cache."
                )
            params = (
                self._params_fn(x, context) if self._params_fn is not None else None
            )
            return self._log_abs_det_jacobian(x, y, params)
        return ladj

    def _log_abs_det_jacobian(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
    ) -> torch.Tensor:
        """
        Computes the log det jacobian `log |dy/dx|` given input and output.
        By default, assumes a volume preserving bijection.
        """

        # TODO: Sum out self.event_dim right-most dimensions
        # self.event_dim may be > 0 for derived classes!
        return torch.zeros_like(x)

    def param_shapes(self, shape: torch.Size) -> Sequence[torch.Size]:
        """
        Abstract method to return shapes of parameters
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    def forward_shape(self, shape: torch.Size) -> torch.Size:
        """
        Infers the shape of the forward computation, given the input shape.
        Defaults to preserving shape.
        """
        return shape

    def inverse_shape(self, shape: torch.Size) -> torch.Size:
        """
        Infers the shapes of the inverse computation, given the output shape.
        Defaults to preserving shape.
        """
        return shape



# Copyright (c) Meta Platforms, Inc

import math
from typing import Optional, Sequence, Tuple

import torch
import torch.distributions.constraints as constraints
import torch.nn.functional as F
from flowtorch.bijectors.fixed import Fixed


class Tanh(Fixed):
    r"""
    Transform via the mapping :math:`y = \tanh(x)`.
    """
    codomain = constraints.interval(-1.0, 1.0)

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        y = torch.tanh(x)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        x = torch.atanh(y)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        return 2.0 * (math.log(2.0) - x - F.softplus(-2.0 * x))



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence, Tuple

import torch
import torch.distributions.constraints as constraints
import torch.nn.functional as F
from flowtorch.bijectors.fixed import Fixed
from flowtorch.ops import eps


class ELU(Fixed):
    codomain = constraints.greater_than(-1.0)

    # TODO: Setting the alpha value of ELU as __init__ argument

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        y = F.elu(x)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        x = torch.max(y, torch.zeros_like(y)) + torch.min(
            torch.log1p(y + eps), torch.zeros_like(y)
        )
        ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        return -F.relu(-x)



# Copyright (c) Meta Platforms, Inc

import math
from typing import Optional, Sequence, Tuple

import flowtorch
import torch
from flowtorch.bijectors.fixed import Fixed
from flowtorch.bijectors.utils import requires_log_detJ


class AffineFixed(Fixed):
    r"""
    Elementwise bijector via the affine mapping :math:`\mathbf{y} = \mu +
    \sigma \otimes \mathbf{x}` where $\mu$ and $\sigma$ are fixed rather than
    learnable.
    """

    # TODO: Handle non-scalar loc and scale with correct broadcasting semantics
    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        loc: float = 0.0,
        scale: float = 1.0
    ) -> None:
        super().__init__(params_fn, shape=shape, context_shape=context_shape)
        self.loc = loc
        self.scale = scale

    def _forward(
        self,
        x: torch.Tensor,
        params: Optional[Sequence[torch.Tensor]],
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        y = self.loc + self.scale * x
        ladj: Optional[torch.Tensor] = None
        if requires_log_detJ():
            ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        x = (y - self.loc) / self.scale
        ladj: Optional[torch.Tensor] = None
        if requires_log_detJ():
            ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        return torch.full_like(x, math.log(abs(self.scale)))



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence

import torch
import torch.distributions
from flowtorch.bijectors.base import Bijector


class VolumePreserving(Bijector):
    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        # TODO: Confirm that this should involve `x`/`self.domain` and not
        # `y`/`self.codomain`
        return torch.zeros(
            x.size()[: -self.domain.event_dim],
            dtype=x.dtype,
            layout=x.layout,  # pyre-ignore[16]
            device=x.device,
        )



# Copyright (c) Meta Platforms, Inc

from typing import Any, Dict, Optional, Union, Iterator

import flowtorch
import torch
import torch.distributions as dist
from torch import Tensor
from torch.distributions.utils import _sum_rightmost
from torch.nn import Parameter


class Flow(torch.nn.Module, dist.Distribution, metaclass=flowtorch.LazyMeta):
    _default_sample_shape = torch.Size()
    arg_constraints: Dict[str, dist.constraints.Constraint] = {}

    def __init__(
        self,
        base_dist: dist.Distribution,
        bijector: flowtorch.Lazy,
        validate_args: Any = None,
    ) -> None:
        torch.nn.Module.__init__(self)

        self.base_dist = base_dist
        self._context: Optional[torch.Tensor] = None
        self.bijector = bijector(shape=base_dist.event_shape)

        # Required so that parameters are registered with nn.Module
        self.params = self.bijector.parameters()  # type: ignore

        # TODO: Confirm that the following logic works. Shouldn't it use
        # .domain and .codomain?? Infer shape from constructed self.bijector
        shape = (
            self.base_dist.batch_shape + self.base_dist.event_shape  # pyre-ignore[16]
        )
        event_dim = self.bijector.domain.event_dim  # type: ignore
        event_dim = max(event_dim, len(self.base_dist.event_shape))
        batch_shape = shape[: len(shape) - event_dim]
        event_shape = shape[len(shape) - event_dim :]

        dist.Distribution.__init__(
            self, batch_shape, event_shape, validate_args=validate_args
        )

    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        for p in super().parameters(recurse=recurse):
            yield p
        if recurse:
            for p in self.bijector.parameters():  # type: ignore
                yield p

    def condition(self, context: torch.Tensor) -> "Flow":
        self._context = context
        return self

    def sample(
        self,
        sample_shape: Union[Tensor, torch.Size] = _default_sample_shape,
        context: Optional[torch.Tensor] = None,
    ) -> Tensor:
        """
        Generates a sample_shape shaped sample or sample_shape shaped batch of
        samples if the distribution parameters are batched. Samples first from
        base distribution and applies `transform()` for every transform in the
        list.
        """
        if context is None:
            context = self._context
        with torch.no_grad():
            x = self.base_dist.sample(sample_shape)
            x = self.bijector.forward(x, context)  # type: ignore
            return x

    def rsample(
        self,
        sample_shape: Union[Tensor, torch.Size] = _default_sample_shape,
        context: Optional[torch.Tensor] = None,
    ) -> Tensor:
        """
        Generates a sample_shape shaped reparameterized sample or sample_shape
        shaped batch of reparameterized samples if the distribution parameters
        are batched. Samples first from base distribution and applies
        `transform()` for every transform in the list.
        """
        if context is None:
            context = self._context
        x = self.base_dist.rsample(sample_shape)
        x = self.bijector.forward(x, context)  # type: ignore
        return x

    def rnormalize(
        self, value: torch.Tensor, context: Optional[torch.Tensor] = None
    ) -> Tensor:
        """
        Push a tensor through the normalizing direction of the flow where
        we can take autodiff gradients on the bijector.
        """
        if context is None:
            context = self._context

        return self.bijector.inverse(value, context)  # type: ignore

    def normalize(
        self, value: torch.Tensor, context: Optional[torch.Tensor] = None
    ) -> Tensor:
        """
        Push a tensor through the normalizing direction of the flow and
        block autodiff gradients on the bijector.
        """
        with torch.no_grad():
            return self.rnormalize(value, context)

    def log_prob(
        self, value: torch.Tensor, context: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Scores the sample by inverting the transform(s) and computing the score
        using the score of the base distribution and the log abs det jacobian.
        """
        if context is None:
            context = self._context
        event_dim = len(self.event_shape)  # pyre-ignore[16]

        x = self.bijector.inverse(value, context)  # type: ignore
        log_prob = -_sum_rightmost(
            self.bijector.log_abs_det_jacobian(x, value, context),  # type: ignore
            event_dim - self.bijector.domain.event_dim,  # type: ignore
        )
        log_prob = log_prob + _sum_rightmost(
            self.base_dist.log_prob(x),
            event_dim - len(self.base_dist.event_shape),  # pyre-ignore[16]
        )

        return log_prob



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence

import torch
from flowtorch import LazyMeta


class Parameters(torch.nn.Module, metaclass=LazyMeta):
    """
    Deferred initialization of parameters.
    """

    def __init__(
        self,
        param_shapes: Sequence[torch.Size],
        input_shape: torch.Size,
        context_shape: Optional[torch.Size],
    ) -> None:
        super().__init__()
        self.input_shape = input_shape
        self.param_shapes = param_shapes
        self.context_shape = context_shape

    def forward(
        self,
        x: Optional[torch.Tensor] = None,
        context: Optional[torch.Tensor] = None,
    ) -> Optional[Sequence[torch.Tensor]]:
        # TODO: Caching etc.
        return self._forward(x, context)

    def _forward(
        self,
        x: Optional[torch.Tensor] = None,
        context: Optional[torch.Tensor] = None,
    ) -> Optional[Sequence[torch.Tensor]]:
        # I raise an exception rather than using @abstractmethod and
        # metaclass=ABC so that we can reserve the metaclass for lazy
        # evaluation.
        raise NotImplementedError()



# Copyright (c) Meta Platforms, Inc

import os

import flowtorch.bijectors as bij
import flowtorch.distributions as dist
import flowtorch.parameters as params
import matplotlib.pyplot as plt
import torch

os.environ["KMP_DUPLICATE_LIB_OK"] = "True"

"""
This is a simple example to demonstrate training of normalizing flows.
Standard bivariate normal noise is sampled from the base distribution
and we learnt to transform it to a bivariate normal distribution with
independent but not identical components (see the produced figures).

"""


def learn_bivariate_normal() -> None:
    # Lazily instantiated flow plus base and target distributions
    bijectors = bij.AffineAutoregressive(
        params_fn=params.DenseAutoregressive(hidden_dims=(32,))
    )
    base_dist = torch.distributions.Independent(
        torch.distributions.Normal(torch.zeros(2), torch.ones(2)), 1
    )
    target_dist = torch.distributions.Independent(
        torch.distributions.Normal(torch.zeros(2) + 5, torch.ones(2) * 0.5), 1
    )

    # Instantiate transformed distribution and parameters
    flow = dist.Flow(base_dist, bijectors)

    # Fixed samples for plotting
    y_initial = flow.sample(
        torch.Size(
            [
                300,
            ]
        )
    )
    y_target = target_dist.sample(
        torch.Size(
            [
                300,
            ]
        )
    )

    # Training loop
    opt = torch.optim.Adam(flow.parameters(), lr=5e-3)
    frame = 0
    for idx in range(3001):
        opt.zero_grad()

        # Minimize KL(p || q)
        y = target_dist.sample((1000,))
        loss = -flow.log_prob(y).mean()

        if idx % 500 == 0:
            print("epoch", idx, "loss", loss)

            # Save SVG
            y_learnt = (
                flow.sample(
                    torch.Size(
                        [
                            300,
                        ]
                    )
                )
                .detach()
                .numpy()
            )

            plt.figure(figsize=(5, 5), dpi=100)
            plt.plot(
                y_target[:, 0],
                y_target[:, 1],
                "o",
                color="blue",
                alpha=0.95,
                label="target",
            )
            plt.plot(
                y_initial[:, 0],
                y_initial[:, 1],
                "o",
                color="grey",
                alpha=0.95,
                label="initial",
            )
            plt.plot(
                y_learnt[:, 0],
                y_learnt[:, 1],
                "o",
                color="red",
                alpha=0.95,
                label="learnt",
            )
            plt.xlim((-4, 8))
            plt.ylim((-4, 8))
            plt.xlabel("$x_1$")
            plt.ylabel("$x_2$")
            plt.legend(loc="lower right", facecolor=(1, 1, 1, 1.0))
            plt.savefig(
                f"bivariate-normal-frame-{frame}.svg",
                bbox_inches="tight",
                transparent=True,
            )

            frame += 1

        loss.backward()
        opt.step()


if __name__ == "__main__":
    learn_bivariate_normal()



# Copyright (c) Meta Platforms, Inc

from typing import Optional

import flowtorch
import torch
from flowtorch.bijectors.elementwise import Elementwise
from flowtorch.bijectors.ops.affine import Affine as AffineOp


class Affine(AffineOp, Elementwise):
    r"""
    Elementwise bijector via the affine mapping :math:`\mathbf{y} = \mu +
    \sigma \otimes \mathbf{x}` where $\mu$ and $\sigma$ are learnable parameters.
    """

    def __init__(
        self,
        params_fn: Optional[flowtorch.Lazy] = None,
        *,
        shape: torch.Size,
        context_shape: Optional[torch.Size] = None,
        log_scale_min_clip: float = -5.0,
        log_scale_max_clip: float = 3.0,
        sigmoid_bias: float = 2.0,
    ) -> None:
        super().__init__(params_fn, shape=shape, context_shape=context_shape)
        self.log_scale_min_clip = log_scale_min_clip
        self.log_scale_max_clip = log_scale_max_clip
        self.sigmoid_bias = sigmoid_bias



# Copyright (c) Meta Platforms, Inc

import importlib
import inspect
import os
import pkgutil
from functools import partial
from typing import Any, Callable, Optional, Sequence, Tuple

import flowtorch
from flowtorch.bijectors.base import Bijector
from flowtorch.parameters.base import Parameters
from torch.distributions import Distribution


copyright_header = """Copyright (c) Meta Platforms, Inc"""


def classname(cls: type) -> str:
    return ".".join([cls.__module__, cls.__name__])


def issubclass_byname(cls: type, test_cls: type) -> bool:
    """
    Test whether a class is a subclass of another by class names, in contrast
    to the built-in issubclass that does it by instance.
    """
    return classname(test_cls) in [classname(c) for c in cls.__mro__]


def isderivedclass(cls: type, base_cls: type) -> bool:
    # NOTE issubclass won't always do what we want here if base_cls is imported
    # inside the module of cls. I.e. issubclass returns False if cls inherits
    # from a base_cls with a different instance.
    return inspect.isclass(cls) and issubclass_byname(cls, base_cls)


def list_bijectors() -> Sequence[Tuple[str, Bijector]]:
    ans = _walk_packages("bijectors", partial(isderivedclass, base_cls=Bijector))
    ans = [a for a in ans if ".ops." not in a[1].__module__]
    return list({classname(cls[1]): cls for cls in ans}.values())


def list_parameters() -> Sequence[Tuple[str, Parameters]]:
    ans = _walk_packages("parameters", partial(isderivedclass, base_cls=Parameters))
    return list({classname(cls[1]): cls for cls in ans}.values())


def list_distributions() -> Sequence[Tuple[str, Parameters]]:
    ans = _walk_packages(
        "distributions", partial(isderivedclass, base_cls=Distribution)
    )
    return list({classname(cls[1]): cls for cls in ans}.values())


def _walk_packages(
    modname: str, filter: Optional[Callable[[Any], bool]]
) -> Sequence[Tuple[str, Any]]:
    classes = []

    # NOTE: I use path of flowtorch rather than e.g. flowtorch.bijectors
    # to avoid circular imports
    path = [os.path.join(flowtorch.__path__[0], modname)]  # type: ignore

    # The followings line uncovered a bug that hasn't been fixed in mypy:
    # https://github.com/python/mypy/issues/1422
    for importer, this_modname, _ in pkgutil.walk_packages(
        path=path,  # type: ignore  # mypy issue #1422
        prefix=f"{flowtorch.__name__}.{modname}.",
        onerror=lambda x: None,
    ):
        # Conditions required for mypy
        if importer is not None:
            if isinstance(importer, importlib.abc.MetaPathFinder):
                finder = importer.find_module(this_modname, None)
            elif isinstance(importer, importlib.abc.PathEntryFinder):
                finder = importer.find_module(this_modname)
        else:
            finder = None

        if finder is not None:
            module = finder.load_module(this_modname)

        else:
            raise Exception("Finder is none")

        if module is not None:
            this_classes = inspect.getmembers(module, filter)
            classes.extend(this_classes)

            del module
            del finder

        else:
            raise Exception("Module is none")

    return classes


class InterfaceError(Exception):
    pass



# Copyright (c) Meta Platforms, Inc

from typing import Sequence, Tuple

import torch
import torch.nn as nn
from torch.nn import functional as F


def sample_mask_indices(
    input_dim: int, hidden_dim: int, simple: bool = True
) -> torch.Tensor:
    """
    Samples the indices assigned to hidden units during the construction of MADE masks
    :param input_dim: the dimensionality of the input variable
    :param hidden_dim: the dimensionality of the hidden layer
    :param simple: True to space fractional indices by rounding to nearest
    int, false round randomly
    """
    indices = torch.linspace(1, input_dim, steps=hidden_dim, device="cpu").to(
        torch.Tensor().device
    )
    if simple:
        # Simple procedure tries to space fractional indices evenly by rounding
        # to nearest int
        return torch.round(indices)
    else:
        # "Non-simple" procedure creates fractional indices evenly then rounds
        # at random
        ints = indices.floor()
        ints += torch.bernoulli(indices - ints)
        return ints


def create_mask(
    input_dim: int,
    context_dim: int,
    hidden_dims: Sequence[int],
    permutation: torch.LongTensor,
    output_multiplier: int,
) -> Tuple[Sequence[torch.Tensor], torch.Tensor]:
    """
    Creates MADE masks for a conditional distribution
    :param input_dim: the dimensionality of the input variable
    :param context_dim: the dimensionality of the variable that is
    conditioned on (for conditional densities)
    :param hidden_dims: the dimensionality of the hidden layers(s)
    :param permutation: the order of the input variables
    :param output_multipliers: tiles the output (e.g. for when a separate
    mean and scale parameter are desired)
    """
    # Create mask indices for input, hidden layers, and final layer
    # We use 0 to refer to the elements of the variable being conditioned on,
    # and range(1:(D_latent+1)) for the input variable
    var_index = torch.empty(permutation.shape, dtype=torch.get_default_dtype())
    var_index[permutation] = torch.arange(input_dim, dtype=torch.get_default_dtype())

    # Create the indices that are assigned to the neurons
    input_indices = torch.cat((torch.zeros(context_dim), 1 + var_index))

    # For conditional MADE, introduce a 0 index that all the conditioned
    # variables are connected to as per Paige and Wood (2016) (see below)
    if context_dim > 0:
        hidden_indices = [sample_mask_indices(input_dim, h) - 1 for h in hidden_dims]
    else:
        hidden_indices = [sample_mask_indices(input_dim - 1, h) for h in hidden_dims]

    # *** TODO: Fix this line ***
    output_indices = (
        (var_index + 1).unsqueeze(-1).repeat(1, output_multiplier).reshape(-1)
    )

    # Create mask from input to output for the skips connections
    mask_skip = (output_indices.unsqueeze(-1) > input_indices.unsqueeze(0)).type_as(
        var_index
    )

    # Create mask from input to first hidden layer, and between subsequent
    # hidden layers
    masks = [
        (hidden_indices[0].unsqueeze(-1) >= input_indices.unsqueeze(0)).type_as(
            var_index
        )
    ]
    for i in range(1, len(hidden_dims)):
        masks.append(
            (
                hidden_indices[i].unsqueeze(-1) >= hidden_indices[i - 1].unsqueeze(0)
            ).type_as(var_index)
        )

    # Create mask from last hidden layer to output layer
    masks.append(
        (output_indices.unsqueeze(-1) > hidden_indices[-1].unsqueeze(0)).type_as(
            var_index
        )
    )

    return masks, mask_skip


class MaskedLinear(nn.Linear):
    """
    A linear mapping with a given mask on the weights (arbitrary bias)
    :param in_features: the number of input features
    :param out_features: the number of output features
    :param mask: the mask to apply to the in_features x out_features weight matrix
    :param bias: whether or not `MaskedLinear` should include a bias term.
    defaults to `True`
    """

    def __init__(
        self, in_features: int, out_features: int, mask: torch.Tensor, bias: bool = True
    ) -> None:
        super().__init__(in_features, out_features, bias)
        self.register_buffer("mask", mask.data)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        masked_weight = self.weight * self.mask
        return F.linear(_input, masked_weight, self.bias)



# Copyright (c) Meta Platforms, Inc

"""
Warning: This file was generated by flowtorch/scripts/generate_imports.py
Do not modify or delete!

"""

from flowtorch.distributions.flow import Flow
from flowtorch.distributions.neals_funnel import NealsFunnel

__all__ = ["Flow", "NealsFunnel"]



# Copyright (c) Meta Platforms, Inc

from typing import Optional, Sequence, Tuple

import flowtorch.ops
import torch
import torch.distributions.constraints as constraints
import torch.nn.functional as F
from flowtorch.bijectors.fixed import Fixed


class Softplus(Fixed):
    r"""
    Elementwise bijector via the mapping :math:`\text{Softplus}(x) = \log(1 + \exp(x))`.
    """
    codomain = constraints.positive

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        y = F.softplus(x)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        x = flowtorch.ops.softplus_inv(y)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        return -F.softplus(-x)



# Copyright (c) Meta Platforms, Inc

import math
from typing import Optional, Sequence, Tuple

import torch
import torch.nn.functional as F
from flowtorch.bijectors.fixed import Fixed


class LeakyReLU(Fixed):
    # TODO: Setting the slope of Leaky ReLU as __init__ argument

    def _forward(
        self, x: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        y = F.leaky_relu(x)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return y, ladj

    def _inverse(
        self, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        x = F.leaky_relu(y, negative_slope=100.0)
        ladj = self._log_abs_det_jacobian(x, y, params)
        return x, ladj

    def _log_abs_det_jacobian(
        self, x: torch.Tensor, y: torch.Tensor, params: Optional[Sequence[torch.Tensor]]
    ) -> torch.Tensor:
        return torch.where(
            x >= 0.0, torch.zeros_like(x), torch.ones_like(x) * math.log(0.01)
        )
